#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2017-07-06 10:38:13
# @Author  : Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.0$
from __future__ import division, print_function, absolute_import
import numpy as np
import numpy.fft as npfft


def freq(fs, n, norm=False, shift=False):
    r"""Return the sample frequencies

    Return the sample frequencies.

    Given a window length `n` and a sample rate `fs`, if shift is ``True``::

      f = [-n/2, ..., n/2] / (d*n)

    Given a window length `n` and a sample rate `fs`, if shift is ``False``::

      f = [0, 1, ..., n] / (d*n)

    If :attr:`norm` is ``True``, :math:`d = 1`, else :math:`d = 1/f_s`.

    Parameters
    ----------
    fs : float
        Sampling rate.
    n : int
        Number of samples.
    norm : bool
        Normalize the frequencies.
    shift : bool
        Does shift the zero frequency to center.

    Returns
    -------
    numpy array
        frequency array with size :math:`n×1`.
    """
    d = 1. / fs

    if shift:
        f = np.linspace(-n / 2., n / 2., n, endpoint=True)
    else:
        f = np.linspace(0, n, n, endpoint=True)
    if norm:
        return f / n
    else:
        return f / (d * n)


def fftfreq(fs, n, norm=False, shift=False):
    r"""Return the Discrete Fourier Transform sample frequencies

    Return the Discrete Fourier Transform sample frequencies.

    Given a window length `n` and a sample rate `fs`, if shift is ``True``::

      f = [-n/2, ..., -1,     0, 1, ...,   n/2-1] / (d*n)   if n is even
      f = [-(n-1)/2, ..., -1, 0, 1, ..., (n-1)/2] / (d*n)   if n is odd

    Given a window length `n` and a sample rate `fs`, if shift is ``False``::

      f = [0, 1, ...,   n/2-1,     -n/2, ..., -1] / (d*n)   if n is even
      f = [0, 1, ..., (n-1)/2, -(n-1)/2, ..., -1] / (d*n)   if n is odd

    where :math:`d = 1/f_s`, if :attr:`norm` is ``True``, :math:`d = 1`, else :math:`d = 1/f_s`.

    Parameters
    ----------
    fs : float
        Sampling rate.
    n : int
        Number of samples.
    norm : bool
        Normalize the frequencies.
    shift : bool
        Does shift the zero frequency to center.

    Returns
    -------
    numpy array
        frequency array with size :math:`n×1`.
    """
    d = 1. / fs
    if n % 2 == 0:
        N = n
        N1 = int(n / 2.)
        N2 = int(n / 2.)
        endpoint = False
    else:
        N = n - 1
        N1 = int((n + 1) / 2.)
        N2 = int((n - 1) / 2.)
        endpoint = True

    if shift:
        f = np.linspace(-N / 2., N / 2., n, endpoint=endpoint)
    else:
        f = np.hstack((np.linspace(0, N / 2., N1, endpoint=endpoint),
                       np.linspace(-N / 2., 0, N2, endpoint=False)))
    if norm:
        return f / n
    else:
        return f / (d * n)


def fftshift(x, axis=None):
    r"""Shift the zero-frequency component to the center of the spectrum.

    This function swaps half-spaces for all axes listed (defaults to all).
    Note that ``y[0]`` is the Nyquist component only if ``len(x)`` is even.

    Parameters
    ----------
    x : numpy array
        The input array.
    axis : int, optional
        Axes over which to shift. (Default is None, which shifts all axes.)

    Returns
    -------
    y : numpy array
        The shifted array.

    See Also
    --------
    ifftshift : The inverse of :func:`fftshift`.

    Examples
    --------

    ::

        import numpy as np
        import pyailib.as ps

        x = [1, 2, 3, 4, 5, 6]
        y = np.fft.fftshift(x)
        print(y)
        y = ps.fftshift(x)
        print(y)

        x = [1, 2, 3, 4, 5, 6, 7]
        y = np.fft.fftshift(x)
        print(y)
        y = ps.fftshift(x)
        print(y)

        axis = (0, 1)  # axis = 0, axis = 1
        x = [[1, 2, 3, 4, 5, 6], [0, 2, 3, 4, 5, 6]]
        y = np.fft.fftshift(x, axis)
        print(y)
        y = ps.fftshift(x, axis)
        print(y)


        x = [[1, 2, 3, 4, 5, 6, 7], [0, 2, 3, 4, 5, 6, 7]]
        y = np.fft.fftshift(x, axis)
        print(y)
        y = ps.fftshift(x, axis)
        print(y)

    """

    if axis is None:
        axis = tuple(range(np.ndim(x)))
    elif type(axis) is int:
        axis = tuple([axis])
    for a in axis:
        n = np.size(x, a)
        p = int(n / 2.)
        x = np.roll(x, p, axis=a)
    return x


def ifftshift(x, axis=None):
    r"""Shift the zero-frequency component back.

    The inverse of :func:`fftshift`. Although identical for even-length `x`, the
    functions differ by one sample for odd-length `x`.

    Parameters
    ----------
    x : numpy array
        The input array.
    axis : int, optional
        Axes over which to shift. (Default is None, which shifts all axes.)

    Returns
    -------
    y : numpy array
        The shifted array.

    See Also
    --------
    fftshift : The inverse of `ifftshift`.

    Examples
    --------

    ::

        x = [1, 2, 3, 4, 5, 6]
        y = np.fft.ifftshift(x)
        print(y)
        y = pl.ifftshift(x)
        print(y)

        x = [1, 2, 3, 4, 5, 6, 7]
        y = np.fft.ifftshift(x)
        print(y)
        y = pl.ifftshift(x)
        print(y)

        axis = (0, 1)  # axis = 0, axis = 1
        x = [[1, 2, 3, 4, 5, 6], [0, 2, 3, 4, 5, 6]]
        y = np.fft.ifftshift(x, axis)
        print(y)
        y = pl.ifftshift(x, axis)
        print(y)


        x = [[1, 2, 3, 4, 5, 6, 7], [0, 2, 3, 4, 5, 6, 7]]
        y = np.fft.ifftshift(x, axis)
        print(y)
        y = pl.ifftshift(x, axis)
        print(y)

    """

    if axis is None:
        axis = tuple(range(np.ndim(x)))
    elif type(axis) is int:
        axis = tuple([axis])
    for a in axis:
        n = np.size(x, a)
        p = int((n + 1) / 2.)
        x = np.roll(x, p, axis=a)
    return x


def padfft(X, nfft=None, axis=0, shift=False):
    r"""PADFT Pad array for doing FFT or IFFT

    PADFT Pad array for doing FFT or IFFT

    Parameters
    ----------
    X : ndarray
        Data to be padded.
    nfft : int or None
        the number of fft point.
    axis : int, optional
        Padding dimension. (the default is 0)
    shift : bool, optional
        Whether to shift the frequency (the default is False)
    """

    if axis is None:
        axis = 0

    Nx = np.size(X, axis)

    if nfft < Nx:
        raise ValueError('Output size is smaller than input size!')

    Nd = np.ndim(X)
    Np = int(np.uint(nfft - Nx))
    PS = np.zeros((Nd, 2), dtype='int32')
    PV = [0]
    if shift:
        PS[axis, 0] = int(np.fix((Np + 1) / 2.))
        X = np.pad(X, PS, 'constant', constant_values=PV)
        PS[axis, :] = [0, Np - PS[axis, 0]]
        X = np.pad(X, PS, 'constant', constant_values=PV)
    else:
        PS[axis, 1] = Np
        X = np.pad(X, PS, 'constant', constant_values=PV)

    return X


def fft(a, n=None, axis=-1, norm=None, shift=False):
    N = np.size(a, axis)
    if (n is not None) and (n > N):
        a = padfft(a, n, axis, shift)
    if shift:
        return npfft.fftshift(npfft.fft(npfft.fftshift(a, axes=axis), n=n, axis=axis, norm=norm), axes=axis)
    else:
        return npfft.fft(a, n=n, axis=axis, norm=norm)


def ifft(a, n=None, axis=-1, norm=None, shift=False):
    if shift:
        return npfft.ifftshift(npfft.ifft(npfft.ifftshift(a, axes=axis), n=n, axis=axis, norm=norm), axes=axis)
    else:
        return npfft.ifft(a, n=n, axis=axis, norm=norm)


def fft2(img):
    r"""
    Improved 2D fft
    """
    out = np.zeros(img.shape, dtype=complex)
    for i in range(out.shape[1]):
        # get range fixed column
        out[:, i] = fft(img[:, i])
    for j in range(out.shape[0]):
        out[j, :] = fft(out[j, :])
    return out


def fftx(x, n=None):

    return npfft.fftshift(npfft.fft(npfft.fftshift(x, n)))


def ffty(x, n=None):
    return (npfft.fftshift(npfft.fft(npfft.fftshift(x.transpose(), n)))).transpose()


def ifftx(x, n=None):
    return npfft.fftshift(npfft.ifft(npfft.fftshift(x), n))


def iffty(x, n=None):
    return (npfft.fftshift(npfft.ifft(x.transpose(), n))).transpose()


if __name__ == "__main__":

    import matplotlib.pyplot as plt

    fs = 1000.
    n = 16

    print(np.fft.fftfreq(n, 1. / fs))

    f = fftfreq(fs, n, shift=False, norm=False)
    print(f)
    f = fftfreq(fs, n, shift=False, norm=True)
    print(f)
    f = fftfreq(fs, n, shift=True, norm=True)
    print(f)

    print(np.linspace(-fs / 2., fs / 2., n))

    Ts = 2.
    f0 = 100.
    Fs = 1000.
    Ns = int(Fs * Ts)
    t = np.linspace(0., Ts, Ns)
    # f = np.linspace(-Fs / 2., Fs / 2., Ns)
    f = fftfreq(Fs, Ns, shift=True)
    print(f)

    x = np.sin(2. * np.pi * f0 * t)
    y = fft(x, shift=True)
    y = np.abs(y)

    plt.figure()
    plt.subplot(121)
    plt.grid()
    plt.plot(t, x)
    plt.subplot(122)
    plt.grid()
    plt.plot(f, y)
    plt.show()

    X = np.array([[1, 2, 3], [4, 5, 6]])
    print(X)
    X = padfft(X, nfft=8, axis=0, shift=False)
    print(X)
    X = padfft(X, nfft=8, axis=1, shift=False)
    print(X)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2018-02-05 16:36:03
# @Author  : Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.0$
from __future__ import division, print_function, absolute_import

import numpy as np
import pyailib as pl


def nextpow2(x):
    r"""get the next higher power of 2.

    Given an number :math:`x`, returns the first p such that :math:`2^p >=|x|`. 

    Args:
        x (int or float): an number.

    Returns:
        int: Next higher power of 2.

    Examples:

        ::

            print(prevpow2(-5), nextpow2(-5))
            print(prevpow2(5), nextpow2(5))
            print(prevpow2(0.3), nextpow2(0.3))
            print(prevpow2(7.3), nextpow2(7.3))
            print(prevpow2(-3.5), nextpow2(-3.5))

            # output
            2 3
            2 3
            -2 -1
            2 3
            1 2

    """

    return int(np.ceil(np.log2(np.abs(x) + 1e-32)))


def prevpow2(x):
    r"""get the previous lower power of 2.

    Given an number :math:`x`, returns the first p such that :math:`2^p <=|x|`. 

    Args:
        x (int or float): an number.

    Returns:
        int: Next higher power of 2.

    Examples:

        ::

            print(prevpow2(-5), nextpow2(-5))
            print(prevpow2(5), nextpow2(5))
            print(prevpow2(0.3), nextpow2(0.3))
            print(prevpow2(7.3), nextpow2(7.3))
            print(prevpow2(-3.5), nextpow2(-3.5))

            # output
            2 3
            2 3
            -2 -1
            2 3
            1 2

    """
    
    return int(np.floor(np.log2(np.abs(x) + 1e-32)))


def ebeo(a, b, op='+'):
    r"""element by element operation

    Element by element operation.

    Parameters
    ----------
    a : list, tuple or ndarray
        The first list/tuple/ndarray.
    b : list, tuple or ndarray
        The second list/tuple/ndarray.
    op : str, optional
        Supported operations are:
        - ``'+'`` or ``'add'`` for addition (default)
        - ``'-'`` or ``'sub'`` for substraction
        - ``'*'`` or ``'mul'`` for multiplication
        - ``'/'`` or ``'div'`` for division
        - ``'**'`` or ``pow`` for power
        - ``'<'``, or ``'lt'`` for less than
        - ``'<='``, or ``'le'`` for less than or equal to
        - ``'>'``, or ``'gt'`` for greater than
        - ``'>='``, or ``'ge'`` for greater than or equal to
        - ``'&'`` for bitwise and
        - ``'|'`` for bitwise or
        - ``'^'`` for bitwise xor
        - function for custom operation.

    Raises
    ------
    TypeError
        If the specified operator not in the above list, raise a TypeError.
    """
    if op in ['+', 'add']:
        return [i + j for i, j in zip(a, b)]
    if op in ['-', 'sub']:
        return [i - j for i, j in zip(a, b)]
    if op in ['*', 'mul']:
        return [i * j for i, j in zip(a, b)]
    if op in ['/', 'div']:
        return [i / j for i, j in zip(a, b)]
    if op in ['**', '^', 'pow']:
        return [i ** j for i, j in zip(a, b)]
    if isinstance(op, str):
        raise TypeError("Not supported operation: " + op + "!")
    else:
        return [op(i, j) for i, j in zip(a, b)]


def r2c(X, caxis=-1, keepdims=False):
    r"""convert real-valued array to complex-valued array

    Convert real-valued array (the size of :attr:`axis` -th dimension is 2) to complex-valued array

    Args:
        X (numpy array): real-valued array.
        caxis (int, optional): the complex axis. Defaults to -1.
        keepdims (bool, optional): keepdims? default is False.

    Returns:
        numpy array: complex-valued array

    Examples:

        ::

            import numpy as np

            np.random.seed(2020)

            Xreal = np.random.randint(0, 30, (3, 2, 4))
            Xcplx = r2c(Xreal, caxis=1)
            Yreal = c2r(Xcplx, caxis=0, keepdims=True)

            print(Xreal, Xreal.shape, 'Xreal')
            print(Xcplx, Xcplx.shape, 'Xcplx')
            print(Yreal, Yreal.shape, 'Yreal')
            print(np.sum(Yreal[0] - Xcplx.real), np.sum(Yreal[1] - Xcplx.imag), 'Error')

            # output
            [[[ 0  8  3 22]
            [ 3 27 29  3]]

            [[ 7 24 29 16]
            [ 0 24 10  9]]

            [[19 11 23 18]
            [ 3  6  5 16]]] (3, 2, 4) Xreal

            [[[ 0. +3.j  8.+27.j  3.+29.j 22. +3.j]]

            [[ 7. +0.j 24.+24.j 29.+10.j 16. +9.j]]

            [[19. +3.j 11. +6.j 23. +5.j 18.+16.j]]] (3, 1, 4) Xcplx

            [[[[ 0.  8.  3. 22.]]

            [[ 7. 24. 29. 16.]]

            [[19. 11. 23. 18.]]]


            [[[ 3. 27. 29.  3.]]

            [[ 0. 24. 10.  9.]]

            [[ 3.  6.  5. 16.]]]] (2, 3, 1, 4) Yreal

            0.0 0.0, Error
    """

    if keepdims:
        idxreal = pl.sl(np.ndim(X), axis=caxis, idx=[[0]])
        idximag = pl.sl(np.ndim(X), axis=caxis, idx=[[1]])
    else:
        idxreal = pl.sl(np.ndim(X), axis=caxis, idx=[0])
        idximag = pl.sl(np.ndim(X), axis=caxis, idx=[1])

    return X[idxreal] + 1j * X[idximag]


def c2r(X, caxis=-1):
    r"""convert complex-valued array to real-valued array

    Args:
        X (numpy array): complex-valued array
        caxis (int, optional): complex axis for real-valued array. Defaults to -1.

    Returns:
        numpy array: real-valued array

    Examples:

        ::

            import numpy as np

            np.random.seed(2020)

            Xreal = np.random.randint(0, 30, (3, 2, 4))
            Xcplx = r2c(Xreal, caxis=1)
            Yreal = c2r(Xcplx, caxis=0, keepdims=True)

            print(Xreal, Xreal.shape, 'Xreal')
            print(Xcplx, Xcplx.shape, 'Xcplx')
            print(Yreal, Yreal.shape, 'Yreal')
            print(np.sum(Yreal[0] - Xcplx.real), np.sum(Yreal[1] - Xcplx.imag), 'Error')

            # output
            [[[ 0  8  3 22]
            [ 3 27 29  3]]

            [[ 7 24 29 16]
            [ 0 24 10  9]]

            [[19 11 23 18]
            [ 3  6  5 16]]] (3, 2, 4) Xreal

            [[[ 0. +3.j  8.+27.j  3.+29.j 22. +3.j]]

            [[ 7. +0.j 24.+24.j 29.+10.j 16. +9.j]]

            [[19. +3.j 11. +6.j 23. +5.j 18.+16.j]]] (3, 1, 4) Xcplx

            [[[[ 0.  8.  3. 22.]]

            [[ 7. 24. 29. 16.]]

            [[19. 11. 23. 18.]]]


            [[[ 3. 27. 29.  3.]]

            [[ 0. 24. 10.  9.]]

            [[ 3.  6.  5. 16.]]]] (2, 3, 1, 4) Yreal

            0.0 0.0, Error
    """

    return np.stack((X.real, X.imag), axis=caxis)


def conj(X, caxis=None):
    r"""conjugates a array

    Both complex and real representation are supported.

    Parameters
    ----------
    X : array
        input
    caxis : int or None
        If :attr:`X` is complex-valued, :attr:`cdim` is ignored. If :attr:`X` is real-valued and :attr:`cdim` is integer
        then :attr:`X` will be treated as complex-valued, in this case, :attr:`cdim` specifies the complex axis;
        otherwise (None), :attr:`X` will be treated as real-valued

    Returns
    -------
    array
         the inputs's conjugate matrix.

    Examples
    ---------

    ::

        np.random.seed(2020)
        X = np.random.rand(2, 3, 3)

        print('---conj')
        print(conj(X, caxis=0))
        print(conj(X[0] + 1j * X[1]))

        # ---output
        ---conj
        [[[ 0.98627683  0.87339195  0.50974552]
        [ 0.27183571  0.33691873  0.21695427]
        [ 0.27647714  0.34331559  0.86215894]]

        [[-0.15669967 -0.14088724 -0.75708028]
        [-0.73632492 -0.35566309 -0.34109302]
        [-0.66680305 -0.21710064 -0.56142698]]]
        [[0.98627683-0.15669967j 0.87339195-0.14088724j 0.50974552-0.75708028j]
        [0.27183571-0.73632492j 0.33691873-0.35566309j 0.21695427-0.34109302j]
        [0.27647714-0.66680305j 0.34331559-0.21710064j 0.86215894-0.56142698j]]

    """

    if np.iscomplex(X).any():  # complex in complex
        return np.conj(X)
    else:
        if caxis is None:  # real
            return X
        else:  # complex in real
            d = np.ndim(X)
            return np.concatenate((X[pl.sl(d, axis=caxis, idx=[[0]])], -X[pl.sl(d, axis=caxis, idx=[[1]])]), axis=caxis)


def real(X, caxis=None, keepdims=False):
    r"""obtain real part of a array

    Both complex and real representation are supported.

    Parameters
    ----------
    X : array
        input
    caxis : int or None
        If :attr:`X` is complex-valued, :attr:`cdim` is ignored. If :attr:`X` is real-valued and :attr:`cdim` is integer
        then :attr:`X` will be treated as complex-valued, in this case, :attr:`cdim` specifies the complex axis;
        otherwise (None), :attr:`X` will be treated as real-valued
    keepdims : bool, optional
        keep dimensions?

    Returns
    -------
    array
         the inputs's real part array.

    Examples
    ---------

    ::

        np.random.seed(2020)
        X = np.random.rand(2, 3, 3)

        print('---real')
        print(real(X, caxis=0))
        print(real(X[0] + 1j * X[1]))

        # ---output
        ---real
        [[0.98627683 0.87339195 0.50974552]
        [0.27183571 0.33691873 0.21695427]
        [0.27647714 0.34331559 0.86215894]]
        [[0.98627683 0.87339195 0.50974552]
        [0.27183571 0.33691873 0.21695427]
        [0.27647714 0.34331559 0.86215894]]
    """

    if np.iscomplex(X).any():  # complex in complex
        return X.real
    else:
        if caxis is None:  # real
            return X
        else:  # complex in real
            d = np.ndim(X)
            idx = [[0]] if keepdims else [0]
            return X[pl.sl(d, axis=caxis, idx=idx)]


def imag(X, caxis=None, keepdims=False):
    r"""obtain imaginary part of a array

    Both complex and real representation are supported.

    Parameters
    ----------
    X : array
        input
    caxis : int or None
        If :attr:`X` is complex-valued, :attr:`cdim` is ignored. If :attr:`X` is real-valued and :attr:`cdim` is integer
        then :attr:`X` will be treated as complex-valued, in this case, :attr:`cdim` specifies the complex axis;
        otherwise (None), :attr:`X` will be treated as real-valued
    keepdims : bool, optional
        keep dimensions?

    Returns
    -------
    array
         the inputs's imaginary part array.

    Examples
    ---------

    ::

        np.random.seed(2020)
        X = np.random.rand(2, 3, 3)

        print('---imag')
        print(imag(X, caxis=0))
        print(imag(X[0] + 1j * X[1]))

        # ---output
        ---imag
        [[0.15669967 0.14088724 0.75708028]
        [0.73632492 0.35566309 0.34109302]
        [0.66680305 0.21710064 0.56142698]]
        [[0.15669967 0.14088724 0.75708028]
        [0.73632492 0.35566309 0.34109302]
        [0.66680305 0.21710064 0.56142698]]

    """

    if np.iscomplex(X).any():  # complex in complex
        return np.imag(X)
    else:
        if caxis is None:  # real
            return np.zeros_like(X)
        else:  # complex in real
            d = np.ndim(X)
            idx = [[1]] if keepdims else [1]
            return X[pl.sl(d, axis=caxis, idx=idx)]


def abs(X, caxis=None, keepdims=False):
    r"""obtain amplitude of a array

    Both complex and real representation are supported.

    .. math::
       {\rm abs}({\bf X}) = |{\bf x}| = \sqrt{u^2 + v^2}, x\in {\bf X}

    where, :math:`u, v` are the real and imaginary part of x, respectively.

    Parameters
    ----------
    X : array
        input
    caxis : int or None
        If :attr:`X` is complex-valued, :attr:`cdim` is ignored. If :attr:`X` is real-valued and :attr:`cdim` is integer
        then :attr:`X` will be treated as complex-valued, in this case, :attr:`cdim` specifies the complex axis;
        otherwise (None), :attr:`X` will be treated as real-valued
    keepdims : bool, optional
        keep dimensions?

    Returns
    -------
    array
         the inputs's amplitude.
   
    Examples
    ---------

    ::

        np.random.seed(2020)
        X = np.random.rand(2, 3, 3)

        print('---abs')
        print(abs(X, caxis=0))
        print(abs(X[0] + 1j * X[1]))

        # ---output
        ---abs
        [[0.99864747 0.88468226 0.91269439]
        [0.78490066 0.48990863 0.40424448]
        [0.72184896 0.40619981 1.02884318]]
        [[0.99864747 0.88468226 0.91269439]
        [0.78490066 0.48990863 0.40424448]
        [0.72184896 0.40619981 1.02884318]]

    """

    if np.iscomplex(X).any():  # complex in complex
        return np.abs(X)
    else:
        if caxis is None:  # real
            return np.abs(X)
        else:  # complex in real
            d = np.ndim(X)
            idxreal = [[0]] if keepdims else [0]
            idximag = [[1]] if keepdims else [1]
            return np.sqrt((X[pl.sl(d, axis=caxis, idx=idxreal)]**2 + X[pl.sl(d, axis=caxis, idx=idximag)]**2))


def pow(X, caxis=None, keepdims=False):
    r"""obtain power of a array

    Both complex and real representation are supported.

    .. math::
       {\rm pow}({\bf X}) = |{\bf x}| = u^2 + v^2, x\in {\bf X}

    where, :math:`u, v` are the real and imaginary part of x, respectively.

    Parameters
    ----------
    X : array
        input
    caxis : int or None
        If :attr:`X` is complex-valued, :attr:`cdim` is ignored. If :attr:`X` is real-valued and :attr:`cdim` is integer
        then :attr:`X` will be treated as complex-valued, in this case, :attr:`cdim` specifies the complex axis;
        otherwise (None), :attr:`X` will be treated as real-valued
    keepdims : bool, optional
        keep dimensions?

    Returns
    -------
    array
         the inputs's power.
   
    Examples
    ---------

    ::

        np.random.seed(2020)
        X = np.random.rand(2, 3, 3)

        print('---pow')
        print(pow(X, caxis=0))
        print(pow(X[0] + 1j * X[1]))

        # ---output
        ---pow
        [[0.99729677 0.78266271 0.83301105]
        [0.61606904 0.24001046 0.1634136 ]
        [0.52106592 0.16499828 1.05851829]]
        [[0.99729677 0.78266271 0.83301105]
        [0.61606904 0.24001046 0.1634136 ]
        [0.52106592 0.16499828 1.05851829]]

    """

    if np.iscomplex(X).any():  # complex in complex
        return (X.conj() * X).real
    else:
        if caxis is None:  # real
            return X**2
        else:  # complex in real
            d = np.ndim(X)
            idxreal = [[0]] if keepdims else [0]
            idximag = [[1]] if keepdims else [1]
            return X[pl.sl(d, axis=caxis, idx=idxreal)]**2 + X[pl.sl(d, axis=caxis, idx=idximag)]**2


if __name__ == '__main__':

    import numpy as np

    np.random.seed(2020)

    Xreal = np.random.randint(0, 30, (3, 2, 4))
    Xcplx = r2c(Xreal, caxis=1, keepdims=True)
    Yreal = c2r(Xcplx, caxis=0)

    print(Xreal, Xreal.shape, 'Xreal')
    print(Xcplx, Xcplx.shape, 'Xcplx')
    print(Yreal, Yreal.shape, 'Yreal')
    print(np.sum(Yreal[0] - Xcplx.real), np.sum(Yreal[1] - Xcplx.imag), 'Error')

    print(prevpow2(-5), nextpow2(-5))
    print(prevpow2(5), nextpow2(5))
    print(prevpow2(0.3), nextpow2(0.3))
    print(prevpow2(7.3), nextpow2(7.3))
    print(prevpow2(-3.5), nextpow2(-3.5))

    np.random.seed(2020)
    X = np.random.rand(2, 3, 3)

    print('---conj')
    print(conj(X, caxis=0))
    print(conj(X[0] + 1j * X[1]))

    print('---real')
    print(real(X, caxis=0))
    print(real(X[0] + 1j * X[1]))

    print('---imag')
    print(imag(X, caxis=0))
    print(imag(X[0] + 1j * X[1]))

    print('---abs')
    print(abs(X, caxis=0))
    print(abs(X[0] + 1j * X[1]))

    print('---pow')
    print(pow(X, caxis=0))
    print(pow(X[0] + 1j * X[1]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2018-02-23 07:01:55
# @Author  : Yan Liu & Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.0$

from __future__ import division, print_function, absolute_import
import h5py
import json
import yaml
import numpy as np
import scipy.io as scio
from pyailib.base.baseops import dreplace


def loadyaml(filepath, field=None):
    """Load a yaml file.

    Parameters
    ----------
    filepath : str
        The file path string.
    field : None, optional
        The string of field that want to be loaded.

    """
    f = open(filepath, 'r', encoding='utf-8')
    if field is None:
        if int(yaml.__version__[0]) > 3:
            data = yaml.load(f, Loader=yaml.FullLoader)
        else:
            data = yaml.load(f)
    else:
        if int(yaml.__version__[0]) > 3:
            data = yaml.load(f, Loader=yaml.FullLoader)[field]
        else:
            data = yaml.load(f)
    return data


def loadjson(filepath, field=None):
    """load a json file

    Parameters
    ----------
    filepath : str
        The file path string.
    field : None, optional
        The string of field that want to be loaded.

    """
    with open(filepath, 'r', encoding='utf-8') as f:
        if field is None:
            data = json.load(f)
        else:
            data = json.load(f)[field]
    return data


def _check_keys(d):
    '''
    checks if entries in dictionary are mat-objects. If yes
    todict is called to change them to nested dictionaries
    '''
    for key in d:
        if isinstance(d[key], scio.matlab.mio5_params.mat_struct):
            d[key] = _todict(d[key])
    return d


def _todict(matobj):
    '''
    A recursive function which constructs from matobjects nested dictionaries
    '''
    d = {}
    for strg in matobj._fieldnames:
        elem = matobj.__dict__[strg]
        if isinstance(elem, scio.matlab.mio5_params.mat_struct):
            d[strg] = _todict(elem)
        else:
            d[strg] = elem
    return d


def loadmat(filepath):
    """load data from an ``.mat`` file

    load data from an ``.mat`` file (:obj:`None` will be replaced by :obj:`None`)

    see https://stackoverflow.com/questions/7008608/scipy-io-loadmat-nested-structures-i-e-dictionaries

    Parameters
    ----------
    filepath : str
        The file path string.

    """
    mdict = scio.loadmat(filepath, struct_as_record=False, squeeze_me=True)
    mdict = _check_keys(mdict)
    dreplace(mdict, fv='None', rv=None, new=False)
    del mdict['__header__'], mdict['__version__'], mdict['__globals__']

    return mdict


def savemat(filepath, mdict, fmt='5'):
    """save data to an ``.mat`` file

    save data to ``.mat`` file (:obj:`None` will be replaced by :obj:`None`)

    Parameters
    ----------
    filepath : str
        savefile path
    mdict : dict
        data in dict formation. 
    fmt : str, optional
        mat formation, by default '5'

    Returns
    -------
    0
        all is ok!
    """
    dreplace(mdict, fv=None, rv='None', new=False)
    scio.savemat(filepath, mdict, format=fmt)
    dreplace(mdict, fv='None', rv=None, new=False)

    return 0


def _create_group_dataset(group, mdict):
    for k, v in mdict.items():
        if k in group.keys():
            del group[k]

        if type(v) is dict:
            subgroup = group.create_group(k)
            _create_group_dataset(subgroup, v)
        else:
            group.create_dataset(k, data=v)


def _read_group_dataset(group, mdict, keys=None):
    if keys is None:
        for k in group.keys():
            if type(group[k]) is h5py.Group:
                exec(k + '={}')
                _read_group_dataset(group[k], eval(k))
                mdict[k] = eval(k)
            else:
                mdict[k] = group[k][()]
    else:
        for k in keys:
            if type(group[k]) is h5py.Group:
                exec(k + '={}')
                _read_group_dataset(group[k], eval(k))
                mdict[k] = eval(k)
            else:
                mdict[k] = group[k][()]


def loadh5(filename, keys=None):
    """load h5 file

    load data from a h5 file. (:obj:`None` will be replaced by :obj:`None`)

    Parameters
    ----------
    filename : str
        File's full path string.
    keys : list
        list of keys.

    Returns
    -------
    D : dict
        The loaded data in ``dict`` type.

    """

    f = h5py.File(filename, 'r')
    D = {}

    _read_group_dataset(f, D, keys)

    dreplace(D, fv='None', rv=None, new=False)

    f.close()
    return D


def saveh5(filename, mdict, mode='w'):
    """save data to h5 file

    save data to h5 file (:obj:`None` will be replaced by :obj:`None`)

    Parameters
    ----------
    filename : str
        filename string
    mdict : dict
        each dict is store in group, the elements in dict are store in dataset
    mode : str
        save mode, ``'w'`` for write, ``'a'`` for add.

    Returns
    -------
    number
        0 --> all is well.
    """

    dreplace(mdict, fv=None, rv='None', new=False)
    f = h5py.File(filename, mode)

    _create_group_dataset(f, mdict)

    f.close()

    dreplace(mdict, fv='None', rv=None, new=False)

    return 0


def mvkeyh5(filepath, ksf, kst, sep='.'):
    """rename keys in ``.h5`` file

    Parameters
    ----------
    filepath : str
        The file path string
    ksf : list
        keys from list, e.g. ['a.x', 'b.y']
    kst : list
        keys to list, e.g. ['a.1', 'b.2']
    sep : str, optional
        The separate pattern, default is ``'.'``

    Returns
    -------
    0
        All is ok!
    """
    ksf = [ksf] if type(ksf) is not list else ksf
    kst = [kst] if type(kst) is not list else kst
    f = h5py.File(filepath, 'a')
    for keyf, keyt in zip(ksf, kst):
        keyf = keyf.split(sep)
        keyt = keyt.split(sep)
        grp = f
        for kf, kt in zip(keyf[:-1], keyt[:-1]):
            grp = grp[kf]
        grp.create_dataset(keyt[-1], data=grp[keyf[-1]][()])
        del grp[keyf[-1]]
    f.close()
    return 0


if __name__ == '__main__':

    a = np.random.randn(3, 4)
    b = 10
    c = [1, 2, 3]
    d = {'d1': 1, 'd2': a}
    s = 'Hello, the future!'
    t = (0, 1)
    n = None

    savemat('./data.mat', {'a': {'x': a, 'y': 1}, 'b': b, 'c': c, 'd': d, 's': s, 'n': n})
    saveh5('./data.h5', {'a': {'x': a}, 'b': b, 'c': c, 'd': d, 's': s, 'n': n})
    data = loadh5('./data.h5', keys=['a', 'd', 's', 'n'])
    for k, v in data.items():
        print(k, v)
    print(data.keys())
    print("==========1")

    data = loadmat('./data.mat')
    for k, v in data.items():
        print(k, v)

    print("==========2")
    # saveh5('./data.h5', {'t': t}, 'w')
    saveh5('./data.h5', {'t': t}, 'a')
    saveh5('./data.h5', {'t': (2, 3, 4)}, 'a')
    data = loadh5('./data.h5')

    for k, v in data.items():
        print(k, v)

    mvkeyh5('./data.h5', ['a.x'], ['a.1'])
    data = loadh5('./data.h5')

    for k, v in data.items():
        print(k, v)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2019-11-07 17:00:48
# @Author  : Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.0$

import numpy as np
import pyailib as pl


def unwrap(x, discont=pl.PI, axis=-1):
    r"""Unwrap by changing deltas between values to :math:`2\pi` complement.

    Unwrap radian phase `x` by changing absoluted jumps greater than
    `discont` to their :math:`2\pi` complement along the given axis.

    Parameters
    ----------
    x : ndarray
        The input.
    discont : float, optional
        Maximum discontinuity between values, default is :math:`\pi`.
    axis : int, optional
        Axis along which unwrap will operate, default is the last axis.

    Returns
    -------
    ndarray
        The unwrapped.
    
    Examples
    --------

    ::

        x = np.array([3.14, -3.12, 3.12, 3.13, -3.11])
        y_np = unwrap(x)
        print(y_np, y_np.shape, type(y_np))

        # output

        tensor([3.1400, 3.1632, 3.1200, 3.1300, 3.1732], dtype=torch.float64) torch.Size([5]) <class 'torch.Tensor'>

    """

    if discont is None:
        discont = pl.PI

    return np.unwrap(x, discont=discont, axis=axis)


def unwrap2(x, discont=pl.PI, axis=-1):
    r"""Unwrap by changing deltas between values to :math:`2\pi` complement.

    Unwrap radian phase `x` by changing absoluted jumps greater than
    `discont` to their :math:`2\pi` complement along the given axis. The elements
    are divided into 2 parts (with equal length) along the given axis.
    The first part is unwrapped in inverse order, while the second part
    is unwrapped in normal order.

    Parameters
    ----------
    x : Tensor
        The input.
    discont : float, optional
        Maximum discontinuity between values, default is :math:`\pi`.
    axis : int, optional
        Axis along which unwrap will operate, default is the last axis.

    Returns
    -------
    Tensor
        The unwrapped.

    see :func:`unwrap`

    Examples
    --------

    ::

        x = np.array([3.14, -3.12, 3.12, 3.13, -3.11])
        y = unwrap(x)
        print(y, y.shape, type(y))

        print("------------------------")
        x = np.array([3.14, -3.12, 3.12, 3.13, -3.11])
        x = np.concatenate((x[::-1], x), axis=0)
        print(x)
        y = unwrap2(x)
        print(y, y.shape, type(y))

        # output
        [3.14       3.16318531 3.12       3.13       3.17318531] (5,) <class 'numpy.ndarray'>
        ------------------------
        [3.17318531 3.13       3.12       3.16318531 3.14       3.14
        3.16318531 3.12       3.13       3.17318531] (10,) <class 'numpy.ndarray'>
    
    """

    d = x.ndim
    s = x.shape[axis]
    i = int(s / 2)
    y1 = np.unwrap(x[pl.sl(d, [axis], [slice(0, i, 1)])][::-1], discont=discont, axis=axis)
    y2 = np.unwrap(x[pl.sl(d, [axis], [slice(i, s, 1)])], discont=discont, axis=axis)
    y = np.concatenate((y1[::-1], y2), axis=axis)
    return y


if __name__ == '__main__':

    x = np.array([3.14, -3.12, 3.12, 3.13, -3.11])
    y = unwrap(x)
    print(y, y.shape, type(y))

    print("------------------------")
    x = np.array([3.14, -3.12, 3.12, 3.13, -3.11])
    x = np.concatenate((x[::-1], x), axis=0)
    y = unwrap2(x)
    print(y, y.shape, type(y))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2018-02-23 07:01:55
# @Author  : Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.0$


from __future__ import division, print_function, absolute_import

import numpy as np
import matplotlib.colors as mcolors
from matplotlib import cm
from pyailib.utils.colormaps import cmaps
from pyailib.base.mathops import nextpow2
from pyailib.misc.transform import scale


# https://sashamaps.net/docs/resources/20-colors/

DISTINCT_COLORS_HEX = {
    'red': '#e6194B', 'green': '#3cb44b', 'yellow': '#ffe119', 'blue': '#4363d8',
    'orange': '#f58231', 'purple': '#911eb4', 'cyan': '#42d4f4', 'magenta': '#f032e6',
    'lime': '#bfef45', 'pink': '#fabed4', 'teal': '#469990', 'lavende': '#dcbeff',
    'brown': '#9A6324', 'beige': '#fffac8', 'maroon': '#800000', 'mint': '#aaffc3',
    'olive': '#808000', 'apricot': '#ffd8b1', 'navy': '#000075', 'grey': '#a9a9a9',
    'white': '#ffffff', 'black': '#000000'
}

DISTINCT_COLORS_RGB = {
    'red': (230, 25, 75), 'green': (60, 180, 75), 'yellow': (255, 225, 25), 'blue': (0, 130, 200),
    'orange': (245, 130, 48), 'purple': (145, 30, 180), 'cyan': (70, 240, 240), 'magenta': (240, 50, 230),
    'lime': (210, 245, 60), 'pink': (250, 190, 212), 'teal': (0, 128, 128), 'lavende': (220, 190, 255),
    'brown': (170, 110, 40), 'beige': (255, 250, 200), 'maroon': (128, 0, 0), 'mint': (170, 255, 195),
    'olive': (128, 128, 0), 'apricot': (255, 215, 180), 'navy': (0, 0, 128), 'grey': (128, 128, 128),
    'white': (255, 255, 255), 'black': (0, 0, 0)
}

DISTINCT_COLORS_CMYK = {
    'red': (0, 100, 66, 0), 'green': (75, 0, 100, 0), 'yellow': (0, 25, 95, 0), 'blue': (100, 35, 0, 0),
    'orange': (0, 60, 92, 0), 'purple': (35, 70, 0, 0), 'cyan': (70, 0, 0, 0), 'magenta': (0, 100, 0, 0),
    'lime': (35, 0, 100, 0), 'pink': (0, 24, 100, 2), 'teal': (100, 0, 0, 50), 'lavende': (14, 25, 100, 0),
    'brown': (0, 35, 75, 33), 'beige': (5, 10, 30, 0), 'maroon': (0, 100, 100, 50), 'mint': (33, 0, 23, 0),
    'olive': (0, 0, 100, 50), 'apricot': (0, 15, 30, 0), 'navy': (100, 100, 0, 50), 'grey': (0, 0, 0, 50),
    'white': (0, 0, 0, 0), 'black': (0, 0, 0, 100)
}

DISTINCT_COLORS_RGB_NORM = {
    'red': (230 / 255, 25 / 255, 75 / 255), 'green': (60 / 255, 180 / 255, 75 / 255), 'yellow': (255 / 255, 225 / 255, 25 / 255), 'blue': (0 / 255, 130 / 255, 200 / 255),
    'orange': (245 / 255, 130 / 255, 48 / 255), 'purple': (145 / 255, 30 / 255, 180 / 255), 'cyan': (70 / 255, 240 / 255, 240 / 255), 'magenta': (240 / 255, 50 / 255, 230 / 255),
    'lime': (210 / 255, 245 / 255, 60 / 255), 'pink': (250 / 255, 190 / 255, 212 / 255), 'teal': (0 / 255, 128 / 255, 128 / 255), 'lavende': (220 / 255, 190 / 255, 255 / 255),
    'brown': (170 / 255, 110 / 255, 40 / 255), 'beige': (255 / 255, 250 / 255, 200 / 255), 'maroon': (128 / 255, 0 / 255, 0 / 255), 'mint': (170 / 255, 255 / 255, 195 / 255),
    'olive': (128 / 255, 128 / 255, 0 / 255), 'apricot': (255 / 255, 215 / 255, 180 / 255), 'navy': (0 / 255, 0 / 255, 128 / 255), 'grey': (128 / 255, 128 / 255, 128 / 255),
    'white': (255 / 255, 255 / 255, 255 / 255), 'black': (0 / 255, 0 / 255, 0 / 255)
}


# https://matplotlib.org/stable/gallery/color/named_colors.html

BASE_COLORS = mcolors.BASE_COLORS
TABLEAU_COLORS = mcolors.TABLEAU_COLORS
CSS4_COLORS = mcolors.CSS4_COLORS


def rgb2gray(rgb, fmt='chnllast'):
    r"""Converts RGB image to GRAY image

    Converts RGB image to GRAY image according to

    .. math::
       G = 0.2989R + 0.5870G + 0.1140B

    see matlab's ``rgb2gray`` function for details.

    Parameters
    ----------
    rgb : numpy array
        Original RGB tensor.
    fmt : str, optional
        Specifies the position of channels in :attr:`rgb` tensor, surpported are:
        - ``'chnllast'`` (default)
        - ``'chnlfirst'``
    """

    dtype = rgb.dtype

    if np.ndim(rgb) < 3:
        return rgb

    if fmt in ['chnllast', 'ChnlLast']:
        return (0.2989 * rgb[..., 0] + 0.5870 * rgb[..., 1] + 0.1140 * rgb[..., 2]).astype(dtype)
    if fmt in ['chnlfirst', 'ChnlFirst']:
        return (0.2989 * rgb[0, ...] + 0.5870 * rgb[1, ...] + 0.1140 * rgb[2, ...]).astype(dtype)


def gray2rgb(gray, cmap, drange=[0, 255], fmtstr=False):
    r"""Converts gray image values to rgb image

    converts `gray` image values to rgb image according to the specified colormap string 'cmap',
    supported are: ``'parula'``, ``'jet'``, ``'hsv'``, ``'hot'``, ``'cool'``, ``'spring'``,
    ``'summer'``, ``'autumn'``, ``'winter'``, ``'gray'``, ``'bone'``, ``'copper'``, ``'pink'``,
    ``'lines'``, ``'colorcube'``, ``'prism'``, ``'flag'``, ``'white'``.

    Parameters
    ----------
    gray : numpy array
        The gray image.
    cmap : str or colormap
        colormap string
    drange : list or tuple
        The low and high value, default is ``[0, 255]``
    fmtstr : bool or str
        Specifies the format (``'int'``, ``'uint8'``, ``'uint16'``, ``'uint32'``, ``'float16'``,
        ``'float32'``, , ``'float64'``) for the output rgb matrix. (default is ``False``, don't change)

    Examples
    --------

    ::

        import pyailib as pl
        import matplotlib.pyplot as plt
        import numpy as np

        cmap = 'jet'
        # cmap = 'hsv'
        # cmap = 'hot'
        # cmap = 'parula'
        gray = pl.imread('../../data/images/LenaGRAY256.png')
        print(gray.shape)

        rgb = gray2rgb(gray, cmap, [0, 1], False)  # rgb --> double, [0, 1]
        # rgb = gray2rgb(gray, cmap, [0, 255], False)  # rgb --> double, [0., 255.]
        # rgb = gray2rgb(gray, cmap, [0, 255], True)  # rgb --> uint8, [0, 255]

        print(gray.shape, np.min(gray), np.max(gray), gray.dtype)
        print(rgb.shape, np.min(rgb), np.max(rgb), rgb.dtype)

        plt.figure()
        plt.subplot(121)
        plt.imshow(gray, cmap=pl.parula if cmap == 'parula' else cmap)
        plt.subplot(122)
        plt.imshow(rgb)
        plt.show()
    """

    N = 256 if (drange[1] == 1) and (drange[0] == 0) else drange[1] - drange[0] + 1

    if cmap == 'parula':
        colormap = cmaps['parula']
        N = 64
    else:
        colormap = cm.get_cmap(cmap, N) if type(cmap) is str else cmap
    gray = scale(gray, st=[0, N], sf=None, istrunc=True, extra=False).astype('int')
    rgb = colormap(gray)[:, :, :3]

    rgb = drange[0] + (drange[1] - drange[0]) * rgb  # [drange(1), drange(2)]
    if fmtstr:
        rgb = rgb.astype(fmtstr)

    return rgb



if __name__ == '__main__':

    A = np.random.randn(5, 4, 3)
    print(A)
    print(rgb2gray(A, fmt='chnlfirst'))

    A = np.random.randn(3, 5, 4)
    print(A)
    print(rgb2gray(A, fmt='chnllast'))

    import pyailib as pl
    import matplotlib.pyplot as plt
    import numpy as np

    cmap = 'jet'
    # cmap = 'hsv'
    # cmap = 'hot'
    # cmap = 'parula'
    gray = pl.imread('../../data/images/LenaGRAY256.png')
    print(gray.shape)

    rgb = gray2rgb(gray, cmap, [0, 1], False)  # rgb --> double, [0, 1]
    # rgb = gray2rgb(gray, cmap, [0, 255], False)  # rgb --> double, [0., 255.]
    # rgb = gray2rgb(gray, cmap, [0, 255], True)  # rgb --> uint8, [0, 255]

    print(gray.shape, np.min(gray), np.max(gray), gray.dtype)
    print(rgb.shape, np.min(rgb), np.max(rgb), rgb.dtype)

    plt.figure()
    plt.subplot(121)
    plt.imshow(gray, cmap=pl.parula if cmap == 'parula' else cmap)
    plt.subplot(122)
    plt.imshow(rgb)
    plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2018-02-05 16:36:03
# @Author  : Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.0$
from __future__ import division, print_function, absolute_import

import numpy as np
import pyailib as pl


def fnorm(X, caxis=None, axis=None, reduction='mean'):
    r"""obtain the f-norm of a array

    Both complex and real representation are supported.

    .. math::
       {\rm fnorm}({\bf X}) = \|{\bf X}\|_2 = \left(\sum_{x_i\in {\bf X}}|x_i|^2\right)^{\frac{1}{2}} = \left(\sum_{x_i\in {\bf X}}(u_i^2 + v_i^2)\right)^{\frac{1}{2}}

    where, :math:`u, v` are the real and imaginary part of :math:`x`, respectively.

    Parameters
    ----------
    X : array
        input
    caxis : int or None
        If :attr:`X` is complex-valued, :attr:`caxis` is ignored. If :attr:`X` is real-valued and :attr:`caxis` is integer
        then :attr:`X` will be treated as complex-valued, in this case, :attr:`caxis` specifies the complex axis;
        otherwise (None), :attr:`X` will be treated as real-valued
    axis : int or None
        The dimension axis (:attr:`caxis` is not included) for computing norm. The default is :obj:`None`, which means all. 
    reduction : str, optional
        The operation in batch dim, :obj:`None`, ``'mean'`` or ``'sum'`` (the default is ``'mean'``)
    
    Returns
    -------
    array
         the inputs's f-norm.

    Examples
    ---------

    ::

        np.random.seed(2020)
        X = np.random.randn(5, 2, 3, 4)

        # real
        C1 = fnorm(X, caxis=None, axis=(-2, -1), reduction=None)
        C2 = fnorm(X, caxis=None, axis=(-2, -1), reduction='sum')
        C3 = fnorm(X, caxis=None, axis=(-2, -1), reduction='mean')
        print(C1, C2, C3)

        # complex in real format
        C1 = fnorm(X, caxis=1, axis=(-2, -1), reduction=None)
        C2 = fnorm(X, caxis=1, axis=(-2, -1), reduction='sum')
        C3 = fnorm(X, caxis=1, axis=(-2, -1), reduction='mean')
        print(C1, C2, C3)

        # complex in complex format
        X = X[:, 0, ...] + 1j * X[:, 1, ...]
        C1 = fnorm(X, caxis=None, axis=(-2, -1), reduction=None)
        C2 = fnorm(X, caxis=None, axis=(-2, -1), reduction='sum')
        C3 = fnorm(X, caxis=None, axis=(-2, -1), reduction='mean')
        print(C1, C2, C3)

        # ---output
        ---norm
        [[3.18214671 3.28727232]
        [3.52423801 3.45821738]
        [3.07757733 3.23720035]
        [2.45488229 3.98372024]
        [2.23480914 3.73551246]] 32.1755762254205 3.21755762254205
        [4.57517398 4.93756225 4.46664844 4.67936684 4.35297889] 23.011730410021634 4.602346082004327
        [4.57517398 4.93756225 4.46664844 4.67936684 4.35297889] 23.011730410021634 4.602346082004327

    """

    if np.iscomplex(X).any():  # complex in complex
        if axis is None:
            F = np.sqrt(np.sum((X.conj() * X).real))
        else:
            F = np.sqrt(np.sum((X.conj() * X).real, axis=axis))
    else:
        if caxis is None:  # real
            if axis is None:
                F = np.sqrt(np.sum(X**2))
            else:
                F = np.sqrt(np.sum(X**2, axis=axis))
        else:  # complex in real
            d = np.ndim(X)
            idxreal = pl.sl(d, axis=caxis, idx=[0])
            idximag = pl.sl(d, axis=caxis, idx=[1])

            if axis is None:
                F = np.sqrt(np.sum(X[idxreal]**2 + X[idximag]**2))
            else:
                F = np.sqrt(np.sum(X[idxreal]**2 + X[idximag]**2, axis=axis))

    if reduction in ['mean', 'MEAN']:
        F = np.mean(F)
    if reduction in ['sum', 'SUM']:
        F = np.sum(F)

    return F

def pnorm(X, caxis=None, axis=None, p=2, reduction='mean'):
    r"""obtain the p-norm of a array

    Both complex and real representation are supported.

    .. math::
       {\rm pnorm}({\bf X}) = \|{\bf X}\|_p = \left(\sum_{x_i\in {\bf X}}|x_i|^p\right)^{\frac{1}{p}} = \left(\sum_{x_i\in {\bf X}}\sqrt{u_i^2+v^2}^p\right)^{\frac{1}{p}}

    where, :math:`u, v` are the real and imaginary part of :math:`x`, respectively.

    Parameters
    ----------
    X : array
        input
    caxis : int or None
        If :attr:`X` is complex-valued, :attr:`caxis` is ignored. If :attr:`X` is real-valued and :attr:`caxis` is integer
        then :attr:`X` will be treated as complex-valued, in this case, :attr:`caxis` specifies the complex axis;
        otherwise (None), :attr:`X` will be treated as real-valued
    axis : int or None
        The dimension axis (:attr:`caxis` is not included) for computing norm. The default is :obj:`None`, which means all. 
    p : int
        Specifies the power. The default is 2.
    reduction : str, optional
        The operation in batch dim, :obj:`None`, ``'mean'`` or ``'sum'`` (the default is ``'mean'``)
    
    Returns
    -------
    array
         the inputs's p-norm.

    Examples
    ---------

    ::

        np.random.seed(2020)
        X = np.random.randn(5, 2, 3, 4)

        # real
        C1 = pnorm(X, caxis=None, axis=(-2, -1), reduction=None)
        C2 = pnorm(X, caxis=None, axis=(-2, -1), reduction='sum')
        C3 = pnorm(X, caxis=None, axis=(-2, -1), reduction='mean')
        print(C1, C2, C3)

        # complex in real format
        C1 = pnorm(X, caxis=1, axis=(-2, -1), reduction=None)
        C2 = pnorm(X, caxis=1, axis=(-2, -1), reduction='sum')
        C3 = pnorm(X, caxis=1, axis=(-2, -1), reduction='mean')
        print(C1, C2, C3)

        # complex in complex format
        X = X[:, 0, ...] + 1j * X[:, 1, ...]
        C1 = pnorm(X, caxis=None, axis=(-2, -1), reduction=None)
        C2 = pnorm(X, caxis=None, axis=(-2, -1), reduction='sum')
        C3 = pnorm(X, caxis=None, axis=(-2, -1), reduction='mean')
        print(C1, C2, C3)

        # ---output
        ---pnorm
        [[3.18214671 3.28727232]
        [3.52423801 3.45821738]
        [3.07757733 3.23720035]
        [2.45488229 3.98372024]
        [2.23480914 3.73551246]] 32.1755762254205 3.21755762254205
        [4.57517398 4.93756225 4.46664844 4.67936684 4.35297889] 23.011730410021634 4.602346082004327
        [4.57517398 4.93756225 4.46664844 4.67936684 4.35297889] 23.011730410021634 4.602346082004327
    """

    if np.iscomplex(X).any():  # complex in complex
        if axis is None:
            F = np.power(np.sum(np.power(np.abs(X), p)), 1/p)
        else:
            F = np.power(np.sum(np.power(np.abs(X), p), axis=axis), 1/p)
    else:
        if caxis is None:  # real
            if axis is None:
                F = np.power(np.sum(np.power(np.abs(X), p)), 1/p)
            else:
                F = np.power(np.sum(np.power(np.abs(X), p), axis=axis), 1/p)
        else:  # complex in real
            d = np.ndim(X)
            idxreal = pl.sl(d, axis=caxis, idx=[0])
            idximag = pl.sl(d, axis=caxis, idx=[1])

            if axis is None:
                F = np.power(np.sum(np.power(np.sqrt(X[idxreal]**2 + X[idximag]**2), p)), 1/p)
            else:
                F = np.power(np.sum(np.power(np.sqrt(X[idxreal]**2 + X[idximag]**2), p), axis=axis), 1/p)

    if reduction in ['mean', 'MEAN']:
        F = np.mean(F)
    if reduction in ['sum', 'SUM']:
        F = np.sum(F)

    return F


if __name__ == '__main__':

    np.random.seed(2020)
    X = np.random.randn(5, 2, 3, 4)

    # real
    C1 = fnorm(X, caxis=None, axis=(-2, -1), reduction=None)
    C2 = fnorm(X, caxis=None, axis=(-2, -1), reduction='sum')
    C3 = fnorm(X, caxis=None, axis=(-2, -1), reduction='mean')
    print(C1, C2, C3)

    # complex in real format
    C1 = fnorm(X, caxis=1, axis=(-2, -1), reduction=None)
    C2 = fnorm(X, caxis=1, axis=(-2, -1), reduction='sum')
    C3 = fnorm(X, caxis=1, axis=(-2, -1), reduction='mean')
    print(C1, C2, C3)

    # complex in complex format
    X = X[:, 0, ...] + 1j * X[:, 1, ...]
    C1 = fnorm(X, caxis=None, axis=(-2, -1), reduction=None)
    C2 = fnorm(X, caxis=None, axis=(-2, -1), reduction='sum')
    C3 = fnorm(X, caxis=None, axis=(-2, -1), reduction='mean')
    print(C1, C2, C3)

    np.random.seed(2020)
    X = np.random.randn(5, 2, 3, 4)

    # real
    C1 = pnorm(X, caxis=None, axis=(-2, -1), reduction=None)
    C2 = pnorm(X, caxis=None, axis=(-2, -1), reduction='sum')
    C3 = pnorm(X, caxis=None, axis=(-2, -1), reduction='mean')
    print(C1, C2, C3)

    # complex in real format
    C1 = pnorm(X, caxis=1, axis=(-2, -1), reduction=None)
    C2 = pnorm(X, caxis=1, axis=(-2, -1), reduction='sum')
    C3 = pnorm(X, caxis=1, axis=(-2, -1), reduction='mean')
    print(C1, C2, C3)

    # complex in complex format
    X = X[:, 0, ...] + 1j * X[:, 1, ...]
    C1 = pnorm(X, caxis=None, axis=(-2, -1), reduction=None)
    C2 = pnorm(X, caxis=None, axis=(-2, -1), reduction='sum')
    C3 = pnorm(X, caxis=None, axis=(-2, -1), reduction='mean')
    print(C1, C2, C3)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2019-07-24 18:29:48
# @Author  : Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.0$

import numpy as np
import pyailib as pl
from pyailib.utils.const import EPS


def entropy(X, caxis=None, axis=None, mode='shannon', reduction='mean'):
    r"""compute the entropy of the inputs

    .. math::
        {\rm ENT} = -\sum_{n=0}^N p_i{\rm log}_2 p_n

    where :math:`N` is the number of pixels, :math:`p_n=\frac{|X_n|^2}{\sum_{n=0}^N|X_n|^2}`.

    Parameters
    ----------
    X : numpy array
        The complex or real inputs, for complex inputs, both complex and real representations are surpported.
    caxis : int or None
        If :attr:`X` is complex-valued, :attr:`caxis` is ignored. If :attr:`X` is real-valued and :attr:`caxis` is integer
        then :attr:`X` will be treated as complex-valued, in this case, :attr:`caxis` specifies the complex axis;
        otherwise (None), :attr:`X` will be treated as real-valued
    axis : int or None
        The dimension axis (:attr:`caxis` is not included) for computing entropy. The default is :obj:`None`, which means all. 
    mode : str, optional
        The entropy mode: ``'shannon'`` or ``'natural'`` (the default is 'shannon')
    reduction : str, optional
        The operation in batch dim, :obj:`None`, ``'mean'`` or ``'sum'`` (the default is ``'mean'``)

    Returns
    -------
    S : scalar or numpy array
        The entropy of the inputs.

    Examples
    ---------

    ::

        np.random.seed(2020)
        X = np.random.randn(5, 2, 3, 4)

        # real
        S1 = entropy(X, caxis=None, axis=(-2, -1), mode='shannon', reduction=None)
        S2 = entropy(X, caxis=None, axis=(-2, -1), mode='shannon', reduction='sum')
        S3 = entropy(X, caxis=None, axis=(-2, -1), mode='shannon', reduction='mean')
        print(S1, S2, S3)

        # complex in real format
        S1 = entropy(X, caxis=1, axis=(-2, -1), mode='shannon', reduction=None)
        S2 = entropy(X, caxis=1, axis=(-2, -1), mode='shannon', reduction='sum')
        S3 = entropy(X, caxis=1, axis=(-2, -1), mode='shannon', reduction='mean')
        print(S1, S2, S3)

        # complex in complex format
        X = X[:, 0, ...] + 1j * X[:, 1, ...]
        S1 = entropy(X, caxis=None, axis=(-2, -1), mode='shannon', reduction=None)
        S2 = entropy(X, caxis=None, axis=(-2, -1), mode='shannon', reduction='sum')
        S3 = entropy(X, caxis=None, axis=(-2, -1), mode='shannon', reduction='mean')
        print(S1, S2, S3)

        # ---output
        [[2.76482544 2.38657794]
        [2.85232291 2.33204624]
        [2.26890769 2.4308547 ]
        [2.50283407 2.56037192]
        [2.76608007 2.47020486]] 25.33502585795305 2.533502585795305
        [3.03089227 2.84108823 2.93389666 3.00868855 2.8229912 ] 14.637556915006513 2.9275113830013026
        [3.03089227 2.84108823 2.93389666 3.00868855 2.8229912 ] 14.637556915006513 2.9275113830013026

    """

    if mode in ['Shannon', 'shannon', 'SHANNON']:
        logfunc = np.log2
    if mode in ['Natural', 'natural', 'NATURAL']:
        logfunc = np.log

    if np.iscomplex(X).any():  # complex in complex
        X = (X * X.conj()).real
    else:
        if caxis is None:  # real
            X = X**2
        else:  # complex in real
            X = np.sum(X**2, axis=caxis)

    P = np.sum(X, axis, keepdims=True)
    p = X / (P + EPS)
    S = -np.sum(p * logfunc(p + EPS), axis)
    if reduction in ['mean', 'MEAN']:
        S = np.mean(S)
    if reduction in ['sum', 'SUM']:
        S = np.sum(S)

    return S


if __name__ == '__main__':

    np.random.seed(2020)
    X = np.random.randn(5, 2, 3, 4)

    # real
    S1 = entropy(X, caxis=None, axis=(-2, -1), mode='shannon', reduction=None)
    S2 = entropy(X, caxis=None, axis=(-2, -1), mode='shannon', reduction='sum')
    S3 = entropy(X, caxis=None, axis=(-2, -1), mode='shannon', reduction='mean')
    print(S1, S2, S3)

    # complex in real format
    S1 = entropy(X, caxis=1, axis=(-2, -1), mode='shannon', reduction=None)
    S2 = entropy(X, caxis=1, axis=(-2, -1), mode='shannon', reduction='sum')
    S3 = entropy(X, caxis=1, axis=(-2, -1), mode='shannon', reduction='mean')
    print(S1, S2, S3)

    # complex in complex format
    X = X[:, 0, ...] + 1j * X[:, 1, ...]
    S1 = entropy(X, caxis=None, axis=(-2, -1), mode='shannon', reduction=None)
    S2 = entropy(X, caxis=None, axis=(-2, -1), mode='shannon', reduction='sum')
    S3 = entropy(X, caxis=None, axis=(-2, -1), mode='shannon', reduction='mean')
    print(S1, S2, S3)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2018-02-05 16:36:03
# @Author  : Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.0$

import numpy as np
from pyailib.utils.const import *

SINC_TABLE_Nq16Ns8 = [
    [-0.003, 0.010, -0.024, 0.062, 0.993, -0.054, 0.021, -0.009],
    [-0.007, 0.021, -0.049, 0.131, 0.973, -0.098, 0.040, -0.017],
    [-0.012, 0.032, -0.075, 0.207, 0.941, -0.134, 0.055, -0.023],
    [-0.016, 0.043, -0.101, 0.287, 0.896, -0.160, 0.066, -0.027],
    [-0.020, 0.054, -0.125, 0.371, 0.841, -0.176, 0.074, -0.030],
    [-0.024, 0.063, -0.147, 0.457, 0.776, -0.185, 0.078, -0.031],
    [-0.027, 0.071, -0.165, 0.542, 0.703, -0.185, 0.079, -0.031],
    [-0.030, 0.076, -0.178, 0.625, 0.625, -0.178, 0.076, -0.030],
    [-0.031, 0.079, -0.185, 0.703, 0.542, -0.165, 0.071, -0.027],
    [-0.031, 0.078, -0.185, 0.776, 0.457, -0.147, 0.063, -0.024],
    [-0.030, 0.074, -0.176, 0.841, 0.371, -0.125, 0.054, -0.020],
    [-0.027, 0.066, -0.160, 0.896, 0.287, -0.101, 0.043, -0.016],
    [-0.023, 0.055, -0.134, 0.941, 0.207, -0.075, 0.032, -0.012],
    [-0.017, 0.040, -0.098, 0.973, 0.131, -0.049, 0.021, -0.007],
    [-0.009, 0.021, -0.054, 0.993, 0.062, -0.024, 0.010, -0.003],
    [-0.000, 0.000, -0.000, 1.000, 0.000, -0.000, 0.000, -0.000],
]


def sinc_table(Nq, Ns):

    X = np.zeros((Nq, Ns))
    # for q in range(Nq, 0, -1):
    #     NN = Ns / (2.0 * q)
    #     X[q - 1, :] = np.linspace(-NN, NN, Ns)

    for q in range(Nq, 0, -1):
        X[q - 1, :] = np.linspace(0, Ns, Ns * q, endpoint=True)[0::q] - Ns / 2.
        # X[q - 1, :] = np.round(X[q - 1, :]) - Ns / 2.
        print(X[q - 1, :])

    kasier_window = np.kaiser(Ns, 2.5).repeat(Nq).reshape(Ns, Nq).transpose()
    # return sinc(X) * kasier_window
    return sinc(X)


def sinc_interp(xin, r=1.0):
    xin = np.array(xin)
    N = xin.size
    M = int(N * r)

    u = np.linspace(0, N, M)
    v = np.linspace(0, N, N)
    xout = []
    for i in u:
        xout.append(np.sum(xin * sinc(v - i)))
    return np.array(xout), np.array(u)


def interp(x, xp, yp, mod='sinc'):
    """interpolation

    interpolation

    Parameters
    ----------
    x : array_like
        The x-coordinates of the interpolated values.

    xp : 1-D sequence of floats
        The x-coordinates of the data points, must be increasing if argument
        `period` is not specified. Otherwise, `xp` is internally sorted after
        normalizing the periodic boundaries with ``xp = xp % period``.

    yp : 1-D sequence of float or complex
        The y-coordinates of the data points, same length as `xp`.

    mod : str, optional
        ``'sinc'`` : sinc interpolation (the default is 'sinc')

    Returns
    -------
    y : float or complex (corresponding to fp) or ndarray
        The interpolated values, same shape as `x`.

    """

    xp = np.array(xp)
    yp = np.array(yp)
    N = yp.size
    M = x.size

    x = np.linspace(0., N, M)
    xp = np.linspace(0., N, N)

    y = []
    for i in x:
        y.append(np.sum(yp * sinc(xp - i)))
    return np.array(y)


def sinc(x):
    x = np.array(x)
    eps = 1.0e-8
    y = np.where(np.abs(PI * x) < eps, 1.0,
                 np.sin(PI * x + eps) / (PI * x + eps))
    return y


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    Ns = 16.0

    t = np.linspace(-Ns / 2., Ns / 2., Ns)
    x = np.sinc(t)
    plt.figure()
    plt.plot(t, x)
    plt.show()

    SINC_TABLE_Nq16Ns8_1 = np.array(SINC_TABLE_Nq16Ns8)
    SINC_TABLE_Nq16Ns8_2 = sinc_table(16, 8)
    plt.figure()
    plt.subplot(121)
    for n in range(16):
        plt.plot(SINC_TABLE_Nq16Ns8_1[n, :])

    plt.subplot(122)
    for n in range(16):
        plt.plot(SINC_TABLE_Nq16Ns8_2[n, :])
    plt.show()

    Ns = 100.0
    T = 64.0
    Fs = Ns / T
    Ts = 1.0 / Fs

    t0 = np.linspace(0, T, Ns)

    a = 0.9
    x0 = t0 * a**t0

    r = 4.0
    t1 = np.linspace(0, T, Ns * r)

    x1 = np.interp(t1, t0, x0)

    x2, _ = sinc_interp(x0, r=r)

    t2 = np.linspace(0, T, Ns * r)
    x2 = interp(t2, t0, x0)

    print(np.sinc(2.3), sinc(2.3), np.sinc(2.3 / np.pi), np.sin(2.3) / 2.3)

    plt.figure()
    plt.plot(t0, x0, '-ob')
    plt.plot(t1, x1, '-g')
    plt.plot(t2, x2, '-r')
    plt.grid()
    plt.xlabel('Time/s')
    plt.ylabel('Amplitude')
    plt.title('interpolation')
    plt.legend(['original', 'np.interp', 'sinc_interp'])

    plt.show()
